#######################################################
#        Internal functions                           #
#######################################################
lik.QMC1.c<-
function(theta,X,y,ni,tabe)
{ .C("lik_QMC1",
     as.double(theta),
     as.integer(length(theta)-1),
     as.double(X),
     as.integer(y[,1]),
      as.integer(y[,2]),
     as.integer(nrow(X)),
     as.integer(ni),
     as.integer(length(ni)),
     as.integer(max(ni)),
     as.double(tabe),
     as.integer(ncol(tabe)),
     out=as.double(0.0), PACKAGE="glmmREML")$out
    }
    
    
grad.QMC1.c<-
function(theta,X,y,ni,tabe)
{ out<-.C("grad_QMC1",
     as.double(theta),
     as.integer(length(theta)-1),
     as.double(X),
     as.integer(y[,1]),
      as.integer(y[,2]),
     as.integer(nrow(X)),
     as.integer(ni),
     as.integer(length(ni)),
     as.integer(max(ni)),
     as.double(tabe),
     as.integer(ncol(tabe)),
     out=as.double(rep(0.0,length(theta)*length(ni))),PACKAGE="glmmREML")$out
  matrix(out, ncol=length(theta))  
    }


grad.tot.QMC1.c<-
function(theta,X,y,ni,tabe)
{ .C("grad_tot_QMC1",
     as.double(theta),
     as.integer(length(theta)-1),
     as.double(X),
     as.integer(y[,1]),
      as.integer(y[,2]),
     as.integer(nrow(X)),
     as.integer(ni),
     as.integer(length(ni)),
     as.integer(max(ni)),
     as.double(tabe),
     as.integer(ncol(tabe)),
     out=as.double(rep(0.0,length(theta))),PACKAGE="glmmREML")$out
      }


lik.GH1.c<-
function(theta,X,y,ni,tabe)
{ .C("lik_GH1",
     as.double(theta),
     as.integer(length(theta)-1),
     as.double(X),
     as.integer(y[,1]),
      as.integer(y[,2]),
     as.integer(nrow(X)),
     as.integer(ni),
     as.integer(length(ni)),
     as.integer(max(ni)),
     as.double(tabe$nodes),
     as.double(tabe$weights),
     as.integer(length(tabe$weights)),
     out=as.double(0.0),PACKAGE="glmmREML")$out
    }


grad.GH1.c<-
function(theta,X,y,ni,tabe)
{ out<- .C("grad_GH1",
     as.double(theta),
     as.integer(length(theta)-1),
     as.double(X),
     as.integer(y[,1]),
      as.integer(y[,2]),
     as.integer(nrow(X)),
     as.integer(ni),
     as.integer(length(ni)),
     as.integer(max(ni)),
     as.double(tabe$nodes),
     as.double(tabe$weights),
     as.integer(length(tabe$weights)),
     out=as.double(rep(0.0,length(theta)*length(ni))),PACKAGE="glmmREML")$out
  matrix(out, ncol=length(theta))  
    }




grad.tot.GH1.c<-
function(theta,X,y,ni,tabe)
{ .C("grad_tot_GH1",
     as.double(theta),
     as.integer(length(theta)-1),
     as.double(X),
     as.integer(y[,1]),
      as.integer(y[,2]),
     as.integer(nrow(X)),
     as.integer(ni),
     as.integer(length(ni)),
     as.integer(max(ni)),
     as.double(tabe$nodes),
     as.double(tabe$weights),
     as.integer(length(tabe$weights)),
     out=as.double(rep(0.0,length(theta))),PACKAGE="glmmREML")$out
      }


#######################################################
#        User-level functions                         #
#######################################################
lik.logit1<- 
function(theta,X,y,ni,mat,type="gl")
{ 
  out<- if(type=="gl") lik.GH1.c(theta,X,y,ni,mat)
       else  lik.QMC1.c(theta,X,y,ni,mat) 
  return(out)
}


grad.logit1<- 
function(theta,X,y,ni,mat,type="gl")
{ 
  out<- if(type=="gl") grad.tot.GH1.c(theta,X,y,ni,mat)
       else  grad.tot.QMC1.c(theta,X,y,ni,mat) 
  return(out)
}


lik.beta.logit1<-
function(beta,psi,X,y,ni,mat,type) 
{ if(type=="gl") lik.GH1.c(c(beta,psi),X,y,ni,mat) 
  else  lik.QMC1.c(c(beta,psi),X,y,ni,mat)
 }


grad.beta.logit1<- 
function(beta,psi,X,y,ni,mat,type)
{
  g<- if(type=="gl") grad.GH1.c(c(beta,psi),X,y,ni,mat) 
       else grad.QMC1.c(c(beta,psi),X,y,ni,mat) 
  out<- as.vector(rowSums(matvec(t(g[,1:length(beta)]),1/g[,length(beta)+1])))
  return(out)
}


prof.logit1<-function(psi,X,y,ni,mat,beta.init,type="gl",method="BFGS")
{ obj<- try(optim(beta.init, lik.beta.logit1, gr=grad.beta.logit1, psi=psi,X=X,ni=ni,mat=mat,y=y, type=type, control=list(fnscale=-1),method=method))
  if(is.numeric(obj$par)) out<- obj$value else out<- -Inf
 return(out)
}


reml.logit1<-function(psi,X,y,ni,mat,beta.init,mle,type="gl",method="BFGS")
{  p<-length(beta.init)
  obj<- try(optim(beta.init, lik.beta.logit1, gr=grad.beta.logit1, psi=psi,X=X,ni=ni,mat=mat,y=y, type=type, control=list(fnscale=-1),method=method,hessian=TRUE))
  g1<- if(type=="gl") grad.GH1.c(c(obj$par,psi),X,y,ni,mat)
       else  grad.QMC1.c(c(obj$par,psi),X,y,ni,mat) 
  g0<- if(type=="gl")  grad.GH1.c(mle,X,y,ni,mat)
      else grad.QMC1.c(mle,X,y,ni,mat) 
 M0<-matvec(t(g0[,1:p]),1/g0[,p+1])
 M1<-matvec(t(g1[,1:p]),1/g1[,p+1])
 S<-M0 %*% t(M1)
 hess<- -obj$hess
 out<-obj$value+0.5*log(det(hess))-log(det(S))
 return(out)
}


