#######################################################
#        Internal functions                           #
#######################################################
lik.QMC2.c<-
function(theta,X,Z,y,ni,tabe)
{ .C("lik_QMC2",
     as.double(theta),
     as.integer(ncol(X)),
     as.integer(length(theta)-ncol(X)),
     as.double(X),
     as.double(Z),
     as.integer(y[,1]),
      as.integer(y[,2]),
     as.integer(nrow(X)),
     as.integer(ni),
     as.integer(length(ni)),
     as.integer(max(ni)),
     as.double(tabe),
     as.integer(dim(tabe)[1]),
     out=as.double(0.0), PACKAGE="glmmREML")$out
    }


grad.QMC2.c<-
function(theta,X,Z,y,ni,tabe)
{ out<- .C("grad_QMC2",
     as.double(theta),
     as.integer(ncol(X)),
     as.integer(length(theta)-ncol(X)),
     as.double(X),
     as.double(Z),
     as.integer(y[,1]),
      as.integer(y[,2]),
     as.integer(nrow(X)),
     as.integer(ni),
     as.integer(length(ni)),
     as.integer(max(ni)),
     as.double(tabe),
     as.integer(dim(tabe)[1]),
     out=as.double(rep(0.0,(ncol(X)+1)*length(ni))), PACKAGE="glmmREML")$out
      matrix(out, ncol=ncol(X)+1)  
    }


lik.GH2.c<-
function(theta,X,Z,y,ni,tabe)
{ .C("lik_GH2",
     as.double(theta),
     as.integer(ncol(X)),
     as.integer(length(theta)-ncol(X)),
     as.double(X),
     as.double(Z),
     as.integer(y[,1]),
      as.integer(y[,2]),
     as.integer(nrow(X)),
     as.integer(ni),
     as.integer(length(ni)),
     as.integer(max(ni)),
     as.double(tabe$nodes),
     as.double(tabe$weights),
     as.integer(length(tabe$weights)),
     out=as.double(0.0), PACKAGE="glmmREML")$out
    }


grad.GH2.c<-
function(theta,X,Z,y,ni,tabe)
{ out<- .C("grad_GH2",
     as.double(theta),
     as.integer(ncol(X)),
     as.integer(length(theta)-ncol(X)),
     as.double(X),
     as.double(Z),
     as.integer(y[,1]),
      as.integer(y[,2]),
     as.integer(nrow(X)),
     as.integer(ni),
     as.integer(length(ni)),
     as.integer(max(ni)),
     as.double(tabe$nodes),
     as.double(tabe$weights),
     as.integer(length(tabe$weights)),
     out=as.double(rep(0.0,(ncol(X)+1)*length(ni))), PACKAGE="glmmREML")$out
      matrix(out, ncol=ncol(X)+1)  
    }

#######################################################
#        User-level functions                         #
#######################################################
# Z matrix n x 2
lik.logit2<- 
function(theta,X,Z,y,ni,mat,type="gl")
{ 
  out<- if(type=="gl") lik.GH2.c(theta,X,Z,y,ni,mat)
       else  lik.QMC2.c(theta,X,Z,y,ni,mat) 
  return(out)
}

lik.beta.logit2<-
function(beta,psi,X,Z,y,ni,mat,type) 
{ if(type=="gl") lik.GH2.c(c(beta,psi),X,Z,y,ni,mat) 
  else  lik.QMC2.c(c(beta,psi),X,Z,y,ni,mat)
 }


grad.beta.logit2<- 
function(beta,psi,X,Z,y,ni,mat,type)
{
  g<- if(type=="gl") grad.GH2.c(c(beta,psi),X,Z,y,ni,mat) 
       else grad.QMC2.c(c(beta,psi),X,Z,y,ni,mat) 
  out<- as.vector(rowSums(matvec(t(g[,1:length(beta)]),1/g[,length(beta)+1])))
  return(out)
}


prof.logit2<-function(psi,X,Z,y,ni,mat,beta.init,type="gl",method="BFGS")
{ obj<- try(optim(beta.init, lik.beta.logit2, gr=grad.beta.logit2, psi=psi,X=X,ni=ni,mat=mat,y=y, 
	Z=Z,type=type, control=list(fnscale=-1),method=method))
  if(is.numeric(obj$par)) out<- obj$value else out<- -Inf
 return(out)
}


reml.logit2<-function(psi,X,Z,y,ni,mat,beta.init,mle,type="gl",method="BFGS")
{  p<-length(beta.init)
  obj<- try(optim(beta.init, lik.beta.logit2, gr=grad.beta.logit2, psi=psi,X=X,ni=ni,mat=mat,y=y, Z=Z,
  type=type, control=list(fnscale=-1),method=method,hessian=TRUE))
  g1<- if(type=="gl") grad.GH2.c(c(obj$par,psi),X,Z,y,ni,mat)
       else  grad.QMC2.c(c(obj$par,psi),X,Z,y,ni,mat) 
  g0<- if(type=="gl")  grad.GH2.c(mle,X,Z,y,ni,mat)
      else grad.QMC2.c(mle,X,Z,y,ni,mat) 
 M0<-matvec(t(g0[,1:p]),1/g0[,p+1])
 M1<-matvec(t(g1[,1:p]),1/g1[,p+1])
 S<-M0 %*% t(M1)
 hess<- -obj$hess
 out<-obj$value+0.5*log(det(hess))-log(det(S))
 return(out)
}


