#######################################################
#        Internal functions                           #
#######################################################
  lik.pois.QMC1.c<-
function(theta,X,y,ni,off,tabe)
{ .C("lik_pois_QMC1",
     as.double(theta),
     as.integer(length(theta)-1),
     as.double(X),
     as.integer(y),
     as.integer(nrow(X)),
     as.integer(ni),
     as.integer(length(ni)),
     as.integer(max(ni)),
     as.double(off),
     as.double(tabe),
     as.integer(ncol(tabe)),
     out=as.double(0.0), PACKAGE="glmmREML")$out
    }


grad.pois.QMC1.c<-
function(theta,X,y,ni,off,tabe)
{ out<-.C("grad_pois_QMC1",
     as.double(theta),
     as.integer(length(theta)-1),
     as.double(X),
     as.integer(y),
     as.integer(nrow(X)),
     as.integer(ni),
     as.integer(length(ni)),
     as.integer(max(ni)),
     as.double(off),
     as.double(tabe),
     as.integer(ncol(tabe)),
     out=as.double(rep(0.0,length(theta)*length(ni))), PACKAGE="glmmREML")$out
  matrix(out, ncol=length(theta))  
    }
  
    

lik.pois.GH1.c<-
function(theta,X,y,ni,off,tabe)
{ .C("lik_pois_GH1",
     as.double(theta),
     as.integer(length(theta)-1),
     as.double(X),
     as.integer(y),
     as.integer(nrow(X)),
     as.integer(ni),
     as.integer(length(ni)),
     as.integer(max(ni)),
     as.double(off),
     as.double(tabe$nodes),
     as.double(tabe$weights),
     as.integer(length(tabe$weights)),
     out=as.double(0.0), PACKAGE="glmmREML")$out
    }




grad.pois.GH1.c<-
function(theta,X,y,ni,off,tabe)
{ out<- .C("grad_pois_GH1",
     as.double(theta),
     as.integer(length(theta)-1),
     as.double(X),
     as.integer(y),
     as.integer(nrow(X)),
     as.integer(ni),
     as.integer(length(ni)),
     as.integer(max(ni)),
     as.double(off),
     as.double(tabe$nodes),
     as.double(tabe$weights),
     as.integer(length(tabe$weights)),
     out=as.double(rep(0.0,length(theta)*length(ni))), PACKAGE="glmmREML")$out
  matrix(out, ncol=length(theta))  
    }



#######################################################
#        User-level functions                         #
#######################################################
lik.pois<- 
function(theta,X,y,ni,offset,mat,type="gl")
{ 
  out<- if(type=="gl") lik.pois.GH1.c(theta,X,y,ni,offset,mat)
       else  lik.pois.QMC1.c(theta,X,y,ni,offset,mat) 
  return(out)
}




lik.beta.pois<-
function(beta,psi,X,y,ni,offset,mat,type) 
{ if(type=="gl") lik.pois.GH1.c(c(beta,psi),X,y,ni,offset,mat) 
  else  lik.pois.QMC1.c(c(beta,psi),X,y,ni,offset,mat)
 }


grad.beta.pois<- 
function(beta,psi,X,y,ni,offset,mat,type)
{
  g<- if(type=="gl") grad.pois.GH1.c(c(beta,psi),X,y,ni,offset,mat) 
       else grad.pois.QMC1.c(c(beta,psi),X,y,ni,offset,mat) 
  out<- as.vector(rowSums(matvec(t(g[,1:length(beta)]),1/g[,length(beta)+1])))
  return(out)
}


prof.pois<-function(psi,X,y,ni,offset,mat,beta.init,type="gl",method="BFGS")
{ obj<- try(optim(beta.init, lik.beta.pois, gr=grad.beta.pois, psi=psi,X=X,ni=ni,mat=mat,y=y, 
	offset=offset,type=type, control=list(fnscale=-1),method=method))
  if(is.numeric(obj$par)) out<- obj$value else out<- -Inf
 return(out)
}


reml.pois<-function(psi,X,y,ni,offset,mat,beta.init,mle,type="gl",method="BFGS")
{ 
	p<-length(beta.init)
  obj<- try(optim(beta.init, lik.beta.pois, gr=grad.beta.pois, psi=psi,X=X,ni=ni,mat=mat,y=y, 
  offset=offset, type=type, control=list(fnscale=-1),method=method,hessian=TRUE))
  g1<- if(type=="gl") grad.pois.GH1.c(c(obj$par,psi),X,y,ni,offset,mat)
       else  grad.pois.QMC1.c(c(obj$par,psi),X,y,ni,offset,mat) 
  g0<- if(type=="gl")  grad.pois.GH1.c(mle,X,y,ni,offset,mat)
      else grad.pois.QMC1.c(mle,X,y,ni,offset,mat) 
 M0<-matvec(t(g0[,1:p]),1/g0[,p+1])
 M1<-matvec(t(g1[,1:p]),1/g1[,p+1])
 S<-M0 %*% t(M1)
 hess<- -obj$hess
 out<-obj$value+0.5*log(det(hess))-log(det(S))
 return(out)
}


