#######################################################
#        Internal functions                           #
#######################################################
lik.GH1.probit.c<-
function(theta,X,y,ni,tabe)
{ .C("lik_GH1_probit",
     as.double(theta),
     as.integer(length(theta)-1),
     as.double(X),
     as.integer(y[,1]),
      as.integer(y[,2]),
     as.integer(nrow(X)),
     as.integer(ni),
     as.integer(length(ni)),
     as.integer(max(ni)),
     as.double(tabe$nodes),
     as.double(tabe$weights),
     as.integer(length(tabe$weights)),
     out=as.double(0.0),PACKAGE="glmmREML")$out
    }


grad.GH1.probit.c<-
function(theta,X,y,ni,tabe)
{ out<- .C("grad_GH1_probit",
     as.double(theta),
     as.integer(length(theta)-1),
     as.double(X),
     as.integer(y[,1]),
      as.integer(y[,2]),
     as.integer(nrow(X)),
     as.integer(ni),
     as.integer(length(ni)),
     as.integer(max(ni)),
     as.double(tabe$nodes),
     as.double(tabe$weights),
     as.integer(length(tabe$weights)),
     out=as.double(rep(0.0,length(theta)*length(ni))),PACKAGE="glmmREML")$out
  matrix(out, ncol=length(theta))  
    }




grad.tot.GH1.probit.c<-
function(theta,X,y,ni,tabe)
{ .C("grad_tot_GH1_probit",
     as.double(theta),
     as.integer(length(theta)-1),
     as.double(X),
     as.integer(y[,1]),
      as.integer(y[,2]),
     as.integer(nrow(X)),
     as.integer(ni),
     as.integer(length(ni)),
     as.integer(max(ni)),
     as.double(tabe$nodes),
     as.double(tabe$weights),
     as.integer(length(tabe$weights)),
     out=as.double(rep(0.0,length(theta))),PACKAGE="glmmREML")$out
      }


#######################################################
#        User-level functions                         #
#######################################################
lik.probit1<- 
function(theta,X,y,ni,mat)
{ 
  out<-lik.GH1.probit.c(theta,X,y,ni,mat)   
  return(out)
}


grad.probit1<- 
function(theta,X,y,ni,mat)
{ 
  out<- grad.tot.GH1.probit.c(theta,X,y,ni,mat)
  return(out)
}


lik.beta.probit1<-
function(beta,psi,X,y,ni,mat) 
{ lik.GH1.probit.c(c(beta,psi),X,y,ni,mat) 
  }


grad.beta.probit1<- 
function(beta,psi,X,y,ni,mat)
{
  g<- grad.GH1.probit.c(c(beta,psi),X,y,ni,mat) 
  out<- as.vector(rowSums(matvec(t(g[,1:length(beta)]),1/g[,length(beta)+1])))
  return(out)
}


prof.probit1<-function(psi,X,y,ni,mat,beta.init,method="BFGS")
{ obj<- try(optim(beta.init, lik.beta.probit1, gr=grad.beta.probit1, psi=psi,X=X,ni=ni,mat=mat,y=y,  control=list(fnscale=-1),method=method))
  if(is.numeric(obj$par)) out<- obj$value else out<- -Inf
 return(out)
}


reml.probit1<-function(psi,X,y,ni,mat,beta.init,mle,method="BFGS")
{ p<-length(beta.init)
  obj<- try(optim(beta.init, lik.beta.probit1, gr=grad.beta.probit1, psi=psi,X=X,ni=ni,mat=mat,y=y,control=list(fnscale=-1),method=method,hessian=TRUE))
  g1<- grad.GH1.probit.c(c(obj$par,psi),X,y,ni,mat)
  g0<- grad.GH1.probit.c(mle,X,y,ni,mat)
  M0<-matvec(t(g0[,1:p]),1/g0[,p+1])
  M1<-matvec(t(g1[,1:p]),1/g1[,p+1])
  S<-M0 %*% t(M1)
  hess<- -obj$hess
  out<-obj$value+0.5*log(det(hess))-log(det(S))
  return(out)
}


