#######################################################
#        Internal functions                           #
#######################################################
lik.tp.GH1.c<-
function(theta,X,Xs,y,ni,tabe)
{ .C("lik_tp_GH1",
     as.double(theta),
     as.integer(length(theta)),
     as.integer(length(theta)-ncol(X)-ncol(Xs)-1),
     as.double(X),
     as.integer(ncol(X)),
     as.double(Xs),
     as.integer(ncol(Xs)),
     as.integer(y[,1]),
     as.double(y[,2]),
     as.integer(nrow(X)),
     as.integer(ni),
     as.integer(length(ni)),
     as.integer(max(ni)),
     as.double(tabe$nodes),
     as.double(tabe$weights),
     as.integer(length(tabe$weights)),
     out=as.double(0.0), PACKAGE="glmmREML")$out
    }


grad.tp.GH1.c<-
function(theta,X,Xs,y,ni,tabe)
{ out<- .C("grad_tp_GH1",
     as.double(theta),
     as.integer(length(theta)),
     as.integer(length(theta)-ncol(X)-ncol(Xs)-1),
     as.double(X),
     as.integer(ncol(X)),
     as.double(Xs),
     as.integer(ncol(Xs)),
     as.integer(y[,1]),
     as.double(y[,2]),
     as.integer(nrow(X)),
     as.integer(ni),
     as.integer(length(ni)),
     as.integer(max(ni)),
     as.double(tabe$nodes),
     as.double(tabe$weights),
     as.integer(length(tabe$weights)),
     out=as.double(rep(0.0,(ncol(X)+ncol(Xs))*length(ni))), PACKAGE="glmmREML")$out
    matrix(out, ncol=ncol(X)+ncol(Xs))  
    }


#######################################################
#        User-level functions                         #
#######################################################
# Z matrix n x 2
lik.tp<- 
function(theta,X,Xs,y,ni,mat,type="gl")
{ 
  out<- if(type=="gl") lik.tp.GH1.c(theta,X,Xs,y,ni,mat)
       else  cat("only Gaussian quadrature available for two-part models") 
  return(out)
}



lik.beta.tp<-
function(beta,psi,X,Xs,y,ni,mat,type) 
{ if(type=="gl") lik.tp.GH1.c(c(beta,psi),X,Xs,y,ni,mat) 
  else  cat("only Gaussian quadrature available for two-part models") 
 }


grad.beta.tp<- 
function(beta,psi,X,Xs,y,ni,mat,type)
{
  g<- if(type=="gl") grad.tp.GH1.c(c(beta,psi),X,Xs,y,ni,mat) 
       else cat("only Gaussian quadrature available for two-part models")
  out<-colSums(g)     
  return(out)
}


prof.tp<-function(psi,X,Xs,y,ni,mat,beta.init,type="gl",method="BFGS")
{ obj<- try(optim(beta.init, lik.beta.tp, gr=grad.beta.tp, psi=psi,X=X,ni=ni,mat=mat,y=y, 
	Xs=Xs,type=type, control=list(fnscale=-1),method=method))
  if(is.numeric(obj$par)) out<- obj$value else out<- -Inf
 return(out)
}


reml.tp<-function(psi,X,Xs,y,ni,mat,beta.init,mle,type="gl",method="BFGS")
{  p<-length(beta.init)
  obj<- try(optim(beta.init, lik.beta.tp, gr=grad.beta.tp, psi=psi,X=X,ni=ni,mat=mat,y=y, Xs=Xs,
  type=type, control=list(fnscale=-1),method=method,hessian=TRUE))
  g1<- if(type=="gl") grad.tp.GH1.c(c(obj$par,psi),X,Xs,y,ni,mat)
       else cat("only Gaussian quadrature available for two-part models")
  g0<- if(type=="gl")  grad.tp.GH1.c(mle,X,Xs,y,ni,mat)
       else cat("only Gaussian quadrature available for two-part models")
  S<- t(g0) %*% g1      
  hess<- -obj$hess
  out<-obj$value+0.5*log(det(hess))-log(det(S))
 return(out)
}


