#include <R.h>
#include <Rmath.h>
#include <R_ext/Applic.h>
#include <stdlib.h>

// auxiliar functions
#define MAT2(mat, i, j, nI) (mat[i+j*nI])
#define MAT3(mat, i, j, k, nI, nJ) (mat[i+j*nI+k*nJ*nI])

void iniz(double *mat,  int le)
{
    int i;
    for (i=0; i<le; i++)  mat[i]=0.0;
}

void matprod(double *a, double *b, int r1, int r2, int c1, int c2, double *c)
{
 int i, j, k; 
 if (c1 != r2){
   Rprintf("Error - Columns of matrix a do not match rows of B\n");
   return;
    }
 for (i=0; i < r1; i++) 
  for (j=0; j < c2; j++)
    MAT2(c,i,j,r1) = 0.0; 

 for (i=0; i < r1; i++) 
  for (j=0; j < c2; j++) 
     for (k=0; k < c1; k++) { MAT2(c,i,j,r1) += MAT2(a,i,k,r1)*MAT2(b,k,j,r2);
    }
}

void trasp(double *mat,
	   int nrow,
	   int ncol,
	   double *tmat)
{
    int i,j;
    for (i=0; i<nrow; i++){
	for (j=0; j<ncol; j++){
	MAT2(tmat,j,i,ncol)=MAT2(mat,i,j,nrow);
	}
    }
}


double eps=2.22e-15;

static double prange(double p)
{
  double mu;
  mu=p;
  
  if (mu<eps)
      mu=eps;
  if (mu>(1-eps))
      mu=(1-eps);
  return(mu);
 }   

/*  #####################################     */
/*         FUNCTIONS FOR LOGIT CASE           */
/* #####################################      */ 

void lik_QMC1(double *theta,
       	     int *dimBeta,
	     double *X,
	     int *y,
	     int *den,
	     int *dimData,
             int *nG,
	     int *dimG,
	     int *maxG, 
	     double *tabe,
	     int *dimQ,
             double *l)
{ double *etab=(double *) R_alloc(*dimData, sizeof(double));
  double *u=(double *) R_alloc(*dimQ, sizeof(double));
  double *csA=(double *) R_alloc(*dimQ, sizeof(double));
  double eta, prob, r;
  int i,j,ind,k,h,n, yM, denM;

  *l=0.0;
  for(i=0; i<*dimData; i++){
    etab[i]=0.0; 
    for(j=0; j< *dimBeta; j++)
      etab[i]+=MAT2(X,i,j,*dimData) * theta[j];}
  ind= 0;
  for(i=0; i<*dimG; i++)
    { n = nG[i];
        
      for(k=0; k<*dimQ; k++)
        u[k]= qnorm(MAT2(tabe,i,k,*dimG),0,1,1,0) * theta[*dimBeta];
      for(h=0; h<*dimQ; h++)
	{csA[h]=0.0;
	  for(j=0; j<n; j++)
	  { k = j + ind;
	    yM =  y[k];      
	    denM = den[k];
            eta = etab[k] + u[h];
	    prob = exp(eta) / (1.0 + exp(eta));
	    (csA[h]) += yM * log(prob)+ (denM - yM) * log(1.0-prob);}
	    } 
      r = 0.0;
      for(h=0; h<*dimQ; h++)
	r += exp(csA[h]) / *dimQ; 
      ind += nG[i];
      (*l) += log(r);
      }
};


void lik_GH1(double *theta,
       	     int *dimBeta,
	     double *X,
	     int *y,
	     int *den,
	     int *dimData,
             int *nG,
	     int *dimG,
	     int *maxG, 
	     double *nod,
	     double *w,
	     int *dimQ,
             double *l)
{ double *etab=(double *) R_alloc(*dimData, sizeof(double));
  double *u=(double *) R_alloc(*dimQ, sizeof(double));
  double *csA=(double *) R_alloc(*dimQ, sizeof(double));
  double eta, prob, r;
  int i,j,ind,k,h,n, yM, denM;
  for(k=0; k<*dimQ; k++)
        u[k]= qnorm((nod[k]+1)/2,0,1,1,0) * theta[*dimBeta];
  *l=0.0;
  for(i=0; i<*dimData; i++){
    etab[i]=0.0; 
    for(j=0; j< *dimBeta; j++)
      etab[i]+=MAT2(X,i,j,*dimData) * theta[j];}
  ind= 0;
  for(i=0; i<*dimG; i++)
    { n = nG[i];
      for(h=0; h<*dimQ; h++)
	{csA[h]=0.0;
	  for(j=0; j<n; j++)
	  { k = j + ind;
	    yM =  y[k];      
	    denM = den[k];
            eta = etab[k] + u[h];
	    prob = exp(eta) / (1.0 + exp(eta));
	    (csA[h]) += yM * log(prob)+ (denM - yM) * log(1.0-prob);}
	    } 
      r = 0.0;
      for(h=0; h<*dimQ; h++)
	r += exp(csA[h]) * w[h] *0.5;
      ind += nG[i];
      (*l) += log(r);
      }
};



void grad_QMC1(double *theta,
       	     int *dimBeta,
	     double *X,
	     int *y,
	     int *den,
	     int *dimData,
             int *nG,
	     int *dimG,
	     int *maxG, 
	     double *tabe,
	     int *dimQ,
             double *grad)
{ double *etab=(double *) R_alloc(*dimData, sizeof(double));
  double *u=(double *) R_alloc(*dimQ, sizeof(double));
  double *csA=(double *) R_alloc(*dimQ, sizeof(double));
  double *csAz=(double *) R_alloc(*dimQ * *dimBeta, sizeof(double));
  double *rp=(double *) R_alloc(*dimBeta, sizeof(double));
  double eta, prob, r;
  int i,j,ind,k,h,n, yM, denM,z;
  
  
  for(i=0; i<*dimData; i++){
    etab[i]=0.0; 
    for(j=0; j< *dimBeta; j++)
      etab[i]+=MAT2(X,i,j,*dimData) * theta[j];}
  ind= 0;
  for(i=0; i<*dimG; i++)
    { n = nG[i];
      for(k=0; k<*dimQ; k++)
        u[k]= qnorm(MAT2(tabe,i,k,*dimG),0,1,1,0) * theta[*dimBeta];
      for(h=0; h<*dimQ; h++)
	{csA[h]=0.0;
         for( z=0; z<*dimBeta; z++)
	   MAT2(csAz, z, h,*dimBeta)=0.0;
	 for(j=0; j<n; j++)
	  { k = j + ind;
	    yM =  y[k];      
	    denM = den[k];
            eta = etab[k] + u[h];
	    prob = exp(eta) / (1.0 + exp(eta));
	    (csA[h]) += yM * log(prob)+ (denM - yM) * log(1.0-prob);
            for( z=0; z<*dimBeta; z++)
              MAT2(csAz,z,h,*dimBeta) += (yM-prob*denM) * MAT2(X,k,z,*dimData) ;}
	}
      r = 0.0;
      iniz(rp, *dimBeta);
      for(h=0; h<*dimQ; h++)
	{ r += exp(csA[h]) / *dimQ;
          for( z=0; z<*dimBeta; z++)
            rp[z] +=  exp(csA[h]) * MAT2(csAz,z,h,*dimBeta) / *dimQ;
	    }
      ind += nG[i];
      MAT2(grad,i,*dimBeta, *dimG) = r;
      for( z=0; z<*dimBeta; z++)
	MAT2(grad,i,z,*dimG) = rp[z];
         }
};



void grad_GH1(double *theta,
       	     int *dimBeta,
	     double *X,
	     int *y,
	     int *den,
	     int *dimData,
             int *nG,
	     int *dimG,
	     int *maxG, 
	     double *nod,
	     double *w,
	     int *dimQ,
             double *grad)
{ double *etab=(double *) R_alloc(*dimData, sizeof(double));
  double *u=(double *) R_alloc(*dimQ, sizeof(double));
  double *csA=(double *) R_alloc(*dimQ, sizeof(double));
  double *csAz=(double *) R_alloc(*dimQ * *dimBeta, sizeof(double));
  double *rp=(double *) R_alloc(*dimBeta, sizeof(double));
  double eta, prob, r;
  int i,j,ind,k,h,n, yM, denM,z;
  
  for(k=0; k<*dimQ; k++)
        u[k]= qnorm((nod[k]+1)/2,0,1,1,0) * theta[*dimBeta];
  for(i=0; i<*dimData; i++){
    etab[i]=0.0; 
    for(j=0; j< *dimBeta; j++)
      etab[i]+=MAT2(X,i,j,*dimData) * theta[j];}
  ind= 0;
  for(i=0; i<*dimG; i++)
    { n = nG[i];
      for(h=0; h<*dimQ; h++)
	{csA[h]=0.0;
         for( z=0; z<*dimBeta; z++)
	   MAT2(csAz, z, h,*dimBeta)=0.0;
	 for(j=0; j<n; j++)
	  { k = j + ind;
	    yM =  y[k];      
	    denM = den[k];
            eta = etab[k] + u[h];
	    prob = exp(eta) / (1.0 + exp(eta));
	    (csA[h]) += yM * log(prob)+ (denM - yM) * log(1.0-prob);
            for( z=0; z<*dimBeta; z++)
              MAT2(csAz,z,h,*dimBeta) += (yM-prob*denM) * MAT2(X,k,z,*dimData) ;}
	}
      r = 0.0;
      iniz(rp, *dimBeta);
      for(h=0; h<*dimQ; h++)
	{ r += exp(csA[h]) * w[h] *0.5;
          for( z=0; z<*dimBeta; z++)
            rp[z] +=  exp(csA[h]) * MAT2(csAz,z,h,*dimBeta) * w[h] * 0.5;	
        }
      ind += nG[i];
      MAT2(grad,i,*dimBeta, *dimG) = r;
      for( z=0; z<*dimBeta; z++)
	MAT2(grad,i,z,*dimG) = rp[z];
         }
};



void grad_tot_GH1(double *theta,
       	     int *dimBeta,
	     double *X,
	     int *y,
	     int *den,
	     int *dimData,
             int *nG,
	     int *dimG,
	     int *maxG, 
	     double *nod,
	     double *w,
	     int *dimQ,
             double *grad)
{ double *etab=(double *) R_alloc(*dimData, sizeof(double));
  double *u=(double *) R_alloc(*dimQ, sizeof(double));
  double *csA=(double *) R_alloc(*dimQ, sizeof(double));
  double *csAz=(double *) R_alloc(*dimQ * (*dimBeta+1), sizeof(double));
  double *rp=(double *) R_alloc(*dimBeta, sizeof(double));
  double eta, prob, r, rs;
  int i,j,ind,k,h,n, yM, denM,z, pp;

  pp = *dimBeta+1;
  for(k=0; k<*dimQ; k++)
        u[k]= qnorm((nod[k]+1)/2,0,1,1,0) * theta[*dimBeta];
  for(i=0; i<*dimData; i++){
    etab[i]=0.0; 
    for(j=0; j< *dimBeta; j++)
      etab[i]+=MAT2(X,i,j,*dimData) * theta[j];}
  ind= 0;
  for(i=0; i<*dimG; i++)
    { n = nG[i];
      for(h=0; h<*dimQ; h++)
	{csA[h]=0.0;
	  for( z=0; z<pp; z++)
	   MAT2(csAz, z, h,pp)=0.0;
	 for(j=0; j<n; j++)
	  { k = j + ind;
	    yM =  y[k];      
	    denM = den[k];
            eta = etab[k] + u[h];
	    prob = exp(eta) / (1.0 + exp(eta));
	    (csA[h]) += yM * log(prob)+ (denM - yM) * log(1.0-prob);
            for( z=0; z<*dimBeta; z++)
              MAT2(csAz,z,h,pp) += (yM-prob*denM) * MAT2(X,k,z,*dimData);
            MAT2(csAz,*dimBeta,h,pp) += (yM-prob*denM) * u[h] / theta[*dimBeta];
	  }
	}
      r = 0.0;
      iniz(rp, *dimBeta);
      rs= 0.0;
      for(h=0; h<*dimQ; h++)
	{ r += exp(csA[h]) * w[h] *0.5;
          for( z=0; z<*dimBeta; z++)
            rp[z] +=  exp(csA[h]) * MAT2(csAz,z,h,pp) * w[h] * 0.5;	
         rs += exp(csA[h]) * MAT2(csAz,*dimBeta,h,pp)*  w[h] * 0.5;
	}
      ind += nG[i];
      for( z=0; z<*dimBeta; z++)
	(grad[z]) += rp[z] / r ;
      grad[*dimBeta] += rs / r;
         }
};


void grad_tot_QMC1(double *theta,
       	     int *dimBeta,
	     double *X,
	     int *y,
	     int *den,
	     int *dimData,
             int *nG,
	     int *dimG,
	     int *maxG, 
	     double *tabe,
	     int *dimQ,
             double *grad)
{ double *etab=(double *) R_alloc(*dimData, sizeof(double));
  double *u=(double *) R_alloc(*dimQ, sizeof(double));
  double *csA=(double *) R_alloc(*dimQ, sizeof(double));
  double *csAz=(double *) R_alloc(*dimQ * (*dimBeta+1), sizeof(double));
  double *rp=(double *) R_alloc(*dimBeta, sizeof(double));
  double eta, prob, r, rs;
  int i,j,ind,k,h,n, yM, denM,z, pp;
  
  pp = *dimBeta+1;
  for(i=0; i<*dimData; i++){
    etab[i]=0.0; 
    for(j=0; j< *dimBeta; j++)
      etab[i]+=MAT2(X,i,j,*dimData) * theta[j];}
  ind= 0;
  for(i=0; i<*dimG; i++)
    { n = nG[i];
      for(k=0; k<*dimQ; k++)
        u[k]= qnorm(MAT2(tabe,i,k,*dimG),0,1,1,0) * theta[*dimBeta];
      for(h=0; h<*dimQ; h++)
	{csA[h]=0.0;
         for( z=0; z<pp; z++)
	   MAT2(csAz, z, h,pp)=0.0;
	 for(j=0; j<n; j++)
	  { k = j + ind;
	    yM =  y[k];      
	    denM = den[k];
            eta = etab[k] + u[h];
	    prob = exp(eta) / (1.0 + exp(eta));
	    (csA[h]) += yM * log(prob)+ (denM - yM) * log(1.0-prob);
            for( z=0; z<*dimBeta; z++)
              MAT2(csAz,z,h,pp) += (yM-prob*denM) * MAT2(X,k,z,*dimData);
            MAT2(csAz,*dimBeta,h,pp) += (yM-prob*denM) * u[h] / theta[*dimBeta];
	  }
	}
      r = 0.0;
      iniz(rp, *dimBeta);
      rs = 0.0;
      for(h=0; h<*dimQ; h++)
	{ r += exp(csA[h]) / *dimQ;
          for( z=0; z<*dimBeta; z++)
            rp[z] +=  exp(csA[h]) * MAT2(csAz,z,h,pp) / *dimQ;
	   rs += exp(csA[h]) * MAT2(csAz,*dimBeta,h,pp) / *dimQ;
	    }
      ind += nG[i];
      for( z=0; z<*dimBeta; z++)
	(grad[z]) += rp[z] / r ;
      grad[*dimBeta] += rs / r;
         }
};


/*  #####################################     */
/*         FUNCTIONS FOR PROBIT CASE          */
/* #####################################      */ 

void lik_GH1_probit(double *theta,
       	     int *dimBeta,
	     double *X,
	     int *y,
	     int *den,
	     int *dimData,
             int *nG,
	     int *dimG,
	     int *maxG, 
	     double *nod,
	     double *w,
	     int *dimQ,
             double *l)
{ double *etab=(double *) R_alloc(*dimData, sizeof(double));
  double *u=(double *) R_alloc(*dimQ, sizeof(double));
  double *csA=(double *) R_alloc(*dimQ, sizeof(double));
  double eta, prob, r;
  int i,j,ind,k,h,n, yM, denM;
  for(k=0; k<*dimQ; k++)
        u[k]= qnorm((nod[k]+1)/2,0,1,1,0) * theta[*dimBeta];
  *l=0.0;
  for(i=0; i<*dimData; i++){
    etab[i]=0.0; 
    for(j=0; j< *dimBeta; j++)
      etab[i]+=MAT2(X,i,j,*dimData) * theta[j];}
  ind= 0;
  for(i=0; i<*dimG; i++)
    { n = nG[i];
      for(h=0; h<*dimQ; h++)
	{csA[h]=0.0;
	  for(j=0; j<n; j++)
	  { k = j + ind;
	    yM =  y[k];      
	    denM = den[k];
            eta = etab[k] + u[h];
	    prob = prange(pnorm(eta,0,1,1,0));
	    (csA[h]) += yM * log(prob)+ (denM - yM) * log(1.0-prob); }
	    } 
      r = 0.0;
      for(h=0; h<*dimQ; h++)
	r += exp(csA[h]) * w[h] *0.5;
      ind += nG[i];
      (*l) += log(r);
      }
};




void grad_GH1_probit(double *theta,
       	     int *dimBeta,
	     double *X,
	     int *y,
	     int *den,
	     int *dimData,
             int *nG,
	     int *dimG,
	     int *maxG, 
	     double *nod,
	     double *w,
	     int *dimQ,
             double *grad)
{ double *etab=(double *) R_alloc(*dimData, sizeof(double));
  double *u=(double *) R_alloc(*dimQ, sizeof(double));
  double *csA=(double *) R_alloc(*dimQ, sizeof(double));
  double *csAz=(double *) R_alloc(*dimQ * *dimBeta, sizeof(double));
  double *rp=(double *) R_alloc(*dimBeta, sizeof(double));
  double eta, prob, r;
  int i,j,ind,k,h,n, yM, denM,z;
  
  for(k=0; k<*dimQ; k++)
        u[k]= qnorm((nod[k]+1)/2,0,1,1,0) * theta[*dimBeta];
  for(i=0; i<*dimData; i++){
    etab[i]=0.0; 
    for(j=0; j< *dimBeta; j++)
      etab[i]+=MAT2(X,i,j,*dimData) * theta[j];}
  ind= 0;
  for(i=0; i<*dimG; i++)
    { n = nG[i];
      for(h=0; h<*dimQ; h++)
	{csA[h]=0.0;
         for( z=0; z<*dimBeta; z++)
	   MAT2(csAz, z, h,*dimBeta)=0.0;
	 for(j=0; j<n; j++)
	  { k = j + ind;
	    yM =  y[k];      
	    denM = den[k];
        eta = etab[k] + u[h];
	    prob = prange(pnorm(eta,0,1,1,0));
	    (csA[h]) += yM * log(prob)+ (denM - yM) * log(1.0-prob);
            for( z=0; z<*dimBeta; z++)
                MAT2(csAz,z,h,*dimBeta) += (yM-prob*denM) * MAT2(X,k,z,*dimData) * dnorm(eta,0,1,0) / (prob * (1-prob)) ;
              }
	}
      r = 0.0;
      iniz(rp, *dimBeta);
      for(h=0; h<*dimQ; h++)
	{ r += exp(csA[h]) * w[h] *0.5;
          for( z=0; z<*dimBeta; z++)
            rp[z] +=  exp(csA[h]) * MAT2(csAz,z,h,*dimBeta) * w[h] * 0.5;	
        }
      ind += nG[i];
      MAT2(grad,i,*dimBeta, *dimG) = r;
      for( z=0; z<*dimBeta; z++)
	MAT2(grad,i,z,*dimG) = rp[z];
         }
};



void grad_tot_GH1_probit(double *theta,
       	     int *dimBeta,
	     double *X,
	     int *y,
	     int *den,
	     int *dimData,
             int *nG,
	     int *dimG,
	     int *maxG, 
	     double *nod,
	     double *w,
	     int *dimQ,
             double *grad)
{ double *etab=(double *) R_alloc(*dimData, sizeof(double));
  double *u=(double *) R_alloc(*dimQ, sizeof(double));
  double *csA=(double *) R_alloc(*dimQ, sizeof(double));
  double *csAz=(double *) R_alloc(*dimQ * (*dimBeta+1), sizeof(double));
  double *rp=(double *) R_alloc(*dimBeta, sizeof(double));
  double eta, prob, r, rs;
  int i,j,ind,k,h,n, yM, denM,z, pp;

  pp = *dimBeta+1;
  for(k=0; k<*dimQ; k++)
        u[k]= qnorm((nod[k]+1)/2,0,1,1,0) * theta[*dimBeta];
  for(i=0; i<*dimData; i++){
    etab[i]=0.0; 
    for(j=0; j< *dimBeta; j++)
      etab[i]+=MAT2(X,i,j,*dimData) * theta[j];}
  ind= 0;
  for(i=0; i<*dimG; i++)
    { n = nG[i];
      for(h=0; h<*dimQ; h++)
	{csA[h]=0.0;
	  for( z=0; z<pp; z++)
	   MAT2(csAz, z, h,pp)=0.0;
	 for(j=0; j<n; j++)
	  { k = j + ind;
	    yM =  y[k];      
	    denM = den[k];
            eta = etab[k] + u[h];
	    prob = prange(pnorm(eta,0,1,1,0));
	    (csA[h]) += yM * log(prob)+ (denM - yM) * log(1.0-prob);
            for( z=0; z<*dimBeta; z++)
              MAT2(csAz,z,h,pp) += (yM-prob*denM) * MAT2(X,k,z,*dimData) * dnorm(eta,0,1,0) / (prob * (1-prob));
            MAT2(csAz,*dimBeta,h,pp) += (yM-prob*denM) * u[h] / theta[*dimBeta] * dnorm(eta,0,1,0) / (prob * (1-prob));
	  }
	}
      r = 0.0;
      iniz(rp, *dimBeta);
      rs= 0.0;
      for(h=0; h<*dimQ; h++)
	{ r += exp(csA[h]) * w[h] *0.5;
          for( z=0; z<*dimBeta; z++)
            rp[z] +=  exp(csA[h]) * MAT2(csAz,z,h,pp) * w[h] * 0.5;	
         rs += exp(csA[h]) * MAT2(csAz,*dimBeta,h,pp)*  w[h] * 0.5;
	}
      ind += nG[i];
      for( z=0; z<*dimBeta; z++)
	(grad[z]) += rp[z] / r ;
      grad[*dimBeta] += rs / r;
         }
};




/*  #####################################     */
/*         FUNCTIONS FOR POISSON CASE           */
/* #####################################      */ 


void lik_pois_QMC1(double *theta,
       	     int *dimBeta,
	     double *X,
	     int *y,
	     int *dimData,
             int *nG,
	     int *dimG,
	     int *maxG, 
	     double *off,
             double *tabe,
	     int *dimQ,
             double *l)
{ double *etab=(double *) R_alloc(*dimData, sizeof(double));
  double *u=(double *) R_alloc(*dimQ, sizeof(double));
  double *csA=(double *) R_alloc(*dimQ, sizeof(double));
  double eta, mu, r;
  int i,j,ind,k,h,n, yM;
  
  *l=0.0;
  for(i=0; i<*dimData; i++){
    etab[i]=off[i]; 
    for(j=0; j< *dimBeta; j++)
      etab[i]+=MAT2(X,i,j,*dimData) * theta[j];}
  ind= 0;
  for(i=0; i<*dimG; i++)
    { n = nG[i];
      //  Rprintf("y: %i, %i, \t", y[i], den[i]);
      for(k=0; k<*dimQ; k++)
        u[k]= qnorm(MAT2(tabe,i,k,*dimG),0,1,1,0) * theta[*dimBeta];
      for(h=0; h<*dimQ; h++)
	{csA[h]=0.0;
	  for(j=0; j<n; j++)
	  { k = j + ind;
	    yM =  y[k];      
	    eta = etab[k] + u[h];
	    mu = exp(eta);
	    (csA[h]) += yM * log(mu)-mu;}
	    } 
      r = 0.0;
      for(h=0; h<*dimQ; h++)
	r += exp(csA[h]) / *dimQ; 
      ind += nG[i];
      (*l) += log(r);
      }
};


void lik_pois_GH1(double *theta,
       	     int *dimBeta,
	     double *X,
	     int *y,
	     int *dimData,
             int *nG,
	     int *dimG,
	     int *maxG, 
	     double *off,
             double *nod,
	     double *w,
	     int *dimQ,
             double *l)
{ double *etab=(double *) R_alloc(*dimData, sizeof(double));
  double *u=(double *) R_alloc(*dimQ, sizeof(double));
  double *csA=(double *) R_alloc(*dimQ, sizeof(double));
  double eta, mu, r;
  int i,j,ind,k,h,n, yM;
  
  for(k=0; k<*dimQ; k++)
        u[k]= qnorm((nod[k]+1)/2,0,1,1,0) * theta[*dimBeta];
  *l=0.0;
  for(i=0; i<*dimData; i++){
    etab[i]=off[i]; 
    for(j=0; j< *dimBeta; j++)
      etab[i]+=MAT2(X,i,j,*dimData) * theta[j];}
  ind= 0;
  for(i=0; i<*dimG; i++)
    { n = nG[i];
      for(h=0; h<*dimQ; h++)
	{csA[h]=0.0;
	  for(j=0; j<n; j++)
	  { k = j + ind;
	    yM =  y[k];      
	    eta = etab[k] + u[h];
	    mu = exp(eta);
	    (csA[h]) += yM * log(mu)-mu;}
	    } 
      r = 0.0;
      for(h=0; h<*dimQ; h++)
	r += exp(csA[h]) * w[h] *0.5;
      ind += nG[i];
      (*l) += log(r);
      }
};


void grad_pois_QMC1(double *theta,
       	     int *dimBeta,
	     double *X,
	     int *y,
	     int *dimData,
             int *nG,
	     int *dimG,
	     int *maxG,
	     double *off,		    
	     double *tabe,
	     int *dimQ,
             double *grad)
{ double *etab=(double *) R_alloc(*dimData, sizeof(double));
  double *u=(double *) R_alloc(*dimQ, sizeof(double));
  double *csA=(double *) R_alloc(*dimQ, sizeof(double));
  double *csAz=(double *) R_alloc(*dimQ * *dimBeta, sizeof(double));
  double *rp=(double *) R_alloc(*dimBeta, sizeof(double));
  double eta, mu, r;
  int i,j,ind,k,h,n, yM,z;
  
  
  for(i=0; i<*dimData; i++){
    etab[i]=off[i]; 
    for(j=0; j< *dimBeta; j++)
      etab[i]+=MAT2(X,i,j,*dimData) * theta[j];}
  ind= 0;
  for(i=0; i<*dimG; i++)
    { n = nG[i];
      for(k=0; k<*dimQ; k++)
        u[k]= qnorm(MAT2(tabe,i,k,*dimG),0,1,1,0) * theta[*dimBeta];
      for(h=0; h<*dimQ; h++)
	{csA[h]=0.0;
         for( z=0; z<*dimBeta; z++)
	   MAT2(csAz, z, h,*dimBeta)=0.0;
	 for(j=0; j<n; j++)
	  { k = j + ind;
	    yM =  y[k];      
	    eta = etab[k] + u[h];
	    mu = exp(eta);
	    (csA[h]) += yM * log(mu) -mu;
            for( z=0; z<*dimBeta; z++)
              MAT2(csAz,z,h,*dimBeta) += (yM-mu) * MAT2(X,k,z,*dimData) ;}
	}
      r = 0.0;
      iniz(rp, *dimBeta);
      for(h=0; h<*dimQ; h++)
	{ r += exp(csA[h]) / *dimQ;
          for( z=0; z<*dimBeta; z++)
            rp[z] +=  exp(csA[h]) * MAT2(csAz,z,h,*dimBeta) / *dimQ;
	    }
      ind += nG[i];
      MAT2(grad,i,*dimBeta, *dimG) = r;
      for( z=0; z<*dimBeta; z++)
	MAT2(grad,i,z,*dimG) = rp[z];
         }
};


void grad_pois_GH1(double *theta,
       	     int *dimBeta,
	     double *X,
	     int *y,
	     int *dimData,
             int *nG,
	     int *dimG,
	     int *maxG,
	     double *off,	   
	     double *nod,
	     double *w,
	     int *dimQ,
             double *grad)
{ double *etab=(double *) R_alloc(*dimData, sizeof(double));
  double *u=(double *) R_alloc(*dimQ, sizeof(double));
  double *csA=(double *) R_alloc(*dimQ, sizeof(double));
  double *csAz=(double *) R_alloc(*dimQ * *dimBeta, sizeof(double));
  double *rp=(double *) R_alloc(*dimBeta, sizeof(double));
  double eta, mu, r;
  int i,j,ind,k,h,n, yM,z;
  
  for(k=0; k<*dimQ; k++)
        u[k]= qnorm((nod[k]+1)/2,0,1,1,0) * theta[*dimBeta];
   for(i=0; i<*dimData; i++){
    etab[i]=off[i]; 
    for(j=0; j< *dimBeta; j++)
      etab[i]+=MAT2(X,i,j,*dimData) * theta[j];}
  ind= 0;
  for(i=0; i<*dimG; i++)
    { n = nG[i];
      for(h=0; h<*dimQ; h++)
	{csA[h]=0.0;
         for( z=0; z<*dimBeta; z++)
	   MAT2(csAz, z, h,*dimBeta)=0.0;
	 for(j=0; j<n; j++)
	  { k = j + ind;
	    yM =  y[k];      
	    eta = etab[k] + u[h];
	    mu = exp(eta);
	    (csA[h]) += yM * log(mu) - mu;
            for( z=0; z<*dimBeta; z++)
              MAT2(csAz,z,h,*dimBeta) += (yM-mu) * MAT2(X,k,z,*dimData) ;}
	}
      r = 0.0;
      iniz(rp, *dimBeta);
      for(h=0; h<*dimQ; h++)
	{ r += exp(csA[h]) * w[h] *0.5;
          for( z=0; z<*dimBeta; z++)
            rp[z] +=  exp(csA[h]) * MAT2(csAz,z,h,*dimBeta) * w[h] * 0.5;	
	    }
      ind += nG[i];
      MAT2(grad,i,*dimBeta, *dimG) = r;
      for( z=0; z<*dimBeta; z++)
	MAT2(grad,i,z,*dimG) = rp[z];
         }
};


/*  #####################################     */
/*         FUNCTIONS FOR NEGBIN CASE           */
/* #####################################      */ 

void lik_nb_QMC1(double *theta,
       	     int *dimBeta,
	     double *X,
	     int *y,
	     int *dimData,
             int *nG,
	     int *dimG,
	     int *maxG, 
	     double *off,
             double *tabe,
	     int *dimQ,
             double *l)
{ double *etab=(double *) R_alloc(*dimData, sizeof(double));
  double *u=(double *) R_alloc(*dimQ, sizeof(double));
  double *csA=(double *) R_alloc(*dimQ, sizeof(double));
  double eta, mu, r,alpha;
  int i,j,ind,k,h,n, yM;
  
  i= *dimBeta +1; 
  alpha= exp(theta[i]);
  *l=0.0;
  for(i=0; i<*dimData; i++){
    etab[i]=off[i]; 
    for(j=0; j< *dimBeta; j++)
      etab[i]+=MAT2(X,i,j,*dimData) * theta[j];}
  ind= 0;
  for(i=0; i<*dimG; i++)
    { n = nG[i];
      //  Rprintf("y: %i, %i, \t", y[i], den[i]);
      for(k=0; k<*dimQ; k++)
        u[k]= qnorm(MAT2(tabe,i,k,*dimG),0,1,1,0) * theta[*dimBeta];
      for(h=0; h<*dimQ; h++)
	{csA[h]=0.0;
	  for(j=0; j<n; j++)
	  { k = j + ind;
	    yM =  y[k];      
	    eta = etab[k] + u[h];
	    mu = exp(eta);
	    (csA[h]) += lgammafn(yM+alpha)+yM*log(mu)-(alpha+yM)*log(alpha+mu)-lgammafn(alpha)+alpha*log(alpha);}
	    } 
      r = 0.0;
      for(h=0; h<*dimQ; h++)
	r += exp(csA[h]) / *dimQ; 
      ind += nG[i];
      (*l) += log(r);
      }
};


void lik_nb_GH1(double *theta,
       	     int *dimBeta,
	     double *X,
	     int *y,
	     int *dimData,
             int *nG,
	     int *dimG,
	     int *maxG, 
	     double *off,
             double *nod,
	     double *w,
	     int *dimQ,
             double *l)
{ double *etab=(double *) R_alloc(*dimData, sizeof(double));
  double *u=(double *) R_alloc(*dimQ, sizeof(double));
  double *csA=(double *) R_alloc(*dimQ, sizeof(double));
  double eta, mu, r,alpha;
  int i,j,ind,k,h,n, yM;
  
  for(k=0; k<*dimQ; k++)
        u[k]= qnorm((nod[k]+1)/2,0,1,1,0) * theta[*dimBeta];
  i= *dimBeta +1; 
  alpha= exp(theta[i]);
  *l=0.0;
  for(i=0; i<*dimData; i++){
    etab[i]=off[i]; 
    for(j=0; j< *dimBeta; j++)
      etab[i]+=MAT2(X,i,j,*dimData) * theta[j];}
  ind= 0;
  for(i=0; i<*dimG; i++)
    { n = nG[i];
      for(h=0; h<*dimQ; h++)
	{csA[h]=0.0;
	  for(j=0; j<n; j++)
	  { k = j + ind;
	    yM =  y[k];      
	    eta = etab[k] + u[h];
	    mu = exp(eta);
	    (csA[h]) +=  lgammafn(yM+alpha)+yM*log(mu)-(alpha+yM)*log(alpha+mu)-lgammafn(alpha)+alpha*log(alpha);}
	    } 
      r = 0.0;
      for(h=0; h<*dimQ; h++)
	r += exp(csA[h]) * w[h] *0.5;
      ind += nG[i];
      (*l) += log(r);
      }
};



void grad_nb_QMC1(double *theta,
       	     int *dimBeta,
	     double *X,
	     int *y,
	     int *dimData,
             int *nG,
	     int *dimG,
	     int *maxG,
	     double *off,		    
	     double *tabe,
	     int *dimQ,
             double *grad)
{ double *etab=(double *) R_alloc(*dimData, sizeof(double));
  double *u=(double *) R_alloc(*dimQ, sizeof(double));
  double *csA=(double *) R_alloc(*dimQ, sizeof(double));
  double *csAz=(double *) R_alloc(*dimQ * *dimBeta, sizeof(double));
  double *rp=(double *) R_alloc(*dimBeta, sizeof(double));
  double eta, mu, r,alpha;
  int i,j,ind,k,h,n, yM,z;

  i= *dimBeta +1; 
  alpha= exp(theta[i]);
  for(i=0; i<*dimData; i++){
    etab[i]=off[i]; 
    for(j=0; j< *dimBeta; j++)
      etab[i]+=MAT2(X,i,j,*dimData) * theta[j];}
  ind= 0;
  for(i=0; i<*dimG; i++)
    { n = nG[i];
      for(k=0; k<*dimQ; k++)
        u[k]= qnorm(MAT2(tabe,i,k,*dimG),0,1,1,0) * theta[*dimBeta];
      for(h=0; h<*dimQ; h++)
	{csA[h]=0.0;
         for( z=0; z<*dimBeta; z++)
	   MAT2(csAz, z, h,*dimBeta)=0.0;
	 for(j=0; j<n; j++)
	  { k = j + ind;
	    yM =  y[k];      
	    eta = etab[k] + u[h];
	    mu = exp(eta);
	    (csA[h]) += lgammafn(yM+alpha)+yM*log(mu)-(alpha+yM)*log(alpha+mu)-lgammafn(alpha)+alpha*log(alpha);
            for( z=0; z<*dimBeta; z++)
              MAT2(csAz,z,h,*dimBeta) += (yM-mu * (alpha+yM) / (alpha+mu) ) * MAT2(X,k,z,*dimData) ;}
	}
      r = 0.0;
      iniz(rp, *dimBeta);
      for(h=0; h<*dimQ; h++)
	{ r += exp(csA[h]) / *dimQ;
          for( z=0; z<*dimBeta; z++)
            rp[z] +=  exp(csA[h]) * MAT2(csAz,z,h,*dimBeta) / *dimQ;
	    }
      ind += nG[i];
      MAT2(grad,i,*dimBeta, *dimG) = r;
      for( z=0; z<*dimBeta; z++)
	MAT2(grad,i,z,*dimG) = rp[z];
         }
};



void grad_nb_GH1(double *theta,
       	     int *dimBeta,
	     double *X,
	     int *y,
	     int *dimData,
             int *nG,
	     int *dimG,
	     int *maxG,
	     double *off,	   
	     double *nod,
	     double *w,
	     int *dimQ,
             double *grad)
{ double *etab=(double *) R_alloc(*dimData, sizeof(double));
  double *u=(double *) R_alloc(*dimQ, sizeof(double));
  double *csA=(double *) R_alloc(*dimQ, sizeof(double));
  double *csAz=(double *) R_alloc(*dimQ * *dimBeta, sizeof(double));
  double *rp=(double *) R_alloc(*dimBeta, sizeof(double));
  double eta, mu, r,alpha;
  int i,j,ind,k,h,n, yM,z;
  
  for(k=0; k<*dimQ; k++)
        u[k]= qnorm((nod[k]+1)/2,0,1,1,0) * theta[*dimBeta];
  i= *dimBeta +1; 
  alpha= exp(theta[i]);
  for(i=0; i<*dimData; i++){
    etab[i]=off[i]; 
    for(j=0; j< *dimBeta; j++)
      etab[i]+=MAT2(X,i,j,*dimData) * theta[j];}
  ind= 0;
  for(i=0; i<*dimG; i++)
    { n = nG[i];
       for(h=0; h<*dimQ; h++)
	{csA[h]=0.0;
         for( z=0; z<*dimBeta; z++)
	   MAT2(csAz, z, h,*dimBeta)=0.0;
	 for(j=0; j<n; j++)
	  { k = j + ind;
	    yM =  y[k];      
	    eta = etab[k] + u[h];
	    mu = exp(eta);
	    (csA[h]) += lgammafn(yM+alpha)+yM*log(mu)-(alpha+yM)*log(alpha+mu)-lgammafn(alpha)+alpha*log(alpha);
            for( z=0; z<*dimBeta; z++)
              MAT2(csAz,z,h,*dimBeta) += (yM-mu * (alpha+yM) / (alpha+mu) ) * MAT2(X,k,z,*dimData) ;}
	}
      r = 0.0;
      iniz(rp, *dimBeta);
      for(h=0; h<*dimQ; h++)
	{ r += exp(csA[h]) * w[h] *0.5;
          for( z=0; z<*dimBeta; z++)
            rp[z] +=  exp(csA[h]) * MAT2(csAz,z,h,*dimBeta) * w[h] * 0.5;	
	    }
      ind += nG[i];
      MAT2(grad,i,*dimBeta, *dimG) = r;
      for( z=0; z<*dimBeta; z++)
	MAT2(grad,i,z,*dimG) = rp[z];
         }
};



/*  #####################################     */
/*         FUNCTIONS FOR LOGIT2 CASE           */
/* #####################################      */ 

void getRt(double *vett, int dimVett, double *Rt)
{ int q, ind, i, j;
  q = (sqrt(1 + 8 * dimVett )-1)/2;
  ind = 0;
  for(i=0; i<q; i++)
   for(j=i; j<q; j++)
     {
       MAT2(Rt,j,i,q) =  vett[ind] ;
      ind += 1;
    }
 }


void lik_QMC2(double *theta,
       	     int *dimBeta,
	     int *dimAlpha,
	     double *X,
	     double *Z, 
	     int *y,
	     int *den,
	     int *dimData,
             int *nG,
	     int *dimG,
	     int *maxG, 
	     double *tabe,
	     int *dimQ,
             double *l)
{ double *etab=(double *) R_alloc(*dimData, sizeof(double));
  double *csA=(double *) R_alloc(*dimQ, sizeof(double));
  double *alpha=(double *) R_alloc(*dimAlpha, sizeof(double));
  double Rt[4] = {0.0, 0.0, 0.0, 0.0};
  double uv[2], uvR[2];
  double eta, prob, r, u;
  int i,j,ind,k,h,n, yM, denM,z;

 for(i=0; i<*dimAlpha; i++)
    { ind = i + *dimBeta;
      alpha[i] = theta[ind];    }
   getRt(alpha, *dimAlpha, Rt);
  *l=0.0;
  for(i=0; i<*dimData; i++){
    etab[i]=0.0; 
    for(j=0; j< *dimBeta; j++)
      etab[i]+=MAT2(X,i,j,*dimData) * theta[j];}
  ind= 0;
  for(i=0; i<*dimG; i++)
    { n = nG[i];
      for(h=0; h<*dimQ; h++)
	{csA[h]=0.0;
         for(z=0; z<2; z++)
	   uv[z]= qnorm(MAT3(tabe,h,i,z,*dimQ,*dimG),0,1,1,0);
	 matprod(Rt,uv,2,2,2,1,uvR);
	 for(j=0; j<n; j++)
	  { k = j + ind;
	    yM =  y[k];      
	    denM = den[k];
	    u = 0.0; 
	    for(z=0; z<2; z++)
	      u += MAT2(Z,k,z,*dimData) * uvR[z];
	     eta = etab[k] + u;
	     prob = exp(eta) / (1.0 + exp(eta));
	    (csA[h]) += yM * log(prob)+ (denM - yM) * log(1.0-prob);}
	    } 
      r = 0.0;
      for(h=0; h<*dimQ; h++)
	r += exp(csA[h]) / *dimQ; 
      ind += nG[i];
      (*l) += log(r);
     
      }
};



void grad_QMC2(double *theta,
       	     int *dimBeta,
	     int *dimAlpha,
	     double *X,
	     double *Z, 
	     int *y,
	     int *den,
	     int *dimData,
             int *nG,
	     int *dimG,
	     int *maxG, 
	     double *tabe,
	     int *dimQ,
             double *grad)
{ double *etab=(double *) R_alloc(*dimData, sizeof(double));
  double *csA=(double *) R_alloc(*dimQ, sizeof(double));
  double *alpha=(double *) R_alloc(*dimAlpha, sizeof(double));
  double *csAz=(double *) R_alloc(*dimQ * *dimBeta, sizeof(double));
  double *rp=(double *) R_alloc(*dimBeta, sizeof(double));
  double Rt[4] = {0.0, 0.0, 0.0, 0.0};
  double uv[2], uvR[2];
  double eta, prob, r, u;
  int i,j,ind,k,h,n, yM, denM,z;

 for(i=0; i<*dimAlpha; i++)
    { ind = i + *dimBeta;
      alpha[i] = theta[ind];    }
   getRt(alpha, *dimAlpha, Rt);
  
  for(i=0; i<*dimData; i++){
    etab[i]=0.0; 
    for(j=0; j< *dimBeta; j++)
      etab[i]+=MAT2(X,i,j,*dimData) * theta[j];}
  ind= 0;
   
  for(i=0; i<*dimG; i++)
    { n = nG[i];
     
      for(h=0; h<*dimQ; h++)
	{csA[h]=0.0;
	 for(z=0; z<2; z++)
	   uv[z]= qnorm(MAT3(tabe,h,i,z,*dimQ,*dimG),0,1,1,0);
	 matprod(Rt,uv,2,2,2,1,uvR);
         for(z=0; z<*dimBeta; z++)
	   MAT2(csAz, z, h,*dimBeta)=0.0;
	 for(j=0; j<n; j++)
	  { k = j + ind;
	    yM =  y[k];      
	    denM = den[k];
	    u = 0.0; 
	    for(z=0; z<2; z++)
	      u += MAT2(Z,k,z,*dimData) * uvR[z];
	    eta = etab[k] + u;
            prob = exp(eta) / (1.0 + exp(eta));
	    (csA[h]) += yM * log(prob)+ (denM - yM) * log(1.0-prob);
            for( z=0; z<*dimBeta; z++)
              MAT2(csAz,z,h,*dimBeta) += (yM-prob*denM) * MAT2(X,k,z,*dimData) ;}
	}
      r = 0.0;
      iniz(rp, *dimBeta);
      for(h=0; h<*dimQ; h++)
	{ r += exp(csA[h]) / *dimQ;
          for( z=0; z<*dimBeta; z++)
            rp[z] +=  exp(csA[h]) * MAT2(csAz,z,h,*dimBeta) / *dimQ;
	    }
      ind += nG[i];
      MAT2(grad,i,*dimBeta, *dimG) = r;
      for( z=0; z<*dimBeta; z++)
	MAT2(grad,i,z,*dimG) = rp[z];
         }
};


   
void lik_GH2(double *theta,
       	     int *dimBeta,
	     int *dimAlpha,
	     double *X,
	     double *Z, 
	     int *y,
	     int *den,
	     int *dimData,
             int *nG,
	     int *dimG,
	     int *maxG,
	     double *nod,
	     double *w,
	     int *dimQ,
             double *l)
{ 
  double *etab=(double *) R_alloc(*dimData, sizeof(double));
  double *alpha=(double *) R_alloc(*dimAlpha, sizeof(double));
  double *v=(double *) R_alloc(*dimQ, sizeof(double));
  double Rt[4] = {0.0, 0.0, 0.0, 0.0};
  double uv[2], uvR[2];
  double eta, prob, r, u, A;
  int i,j,ind,k,h,n, yM, denM,z, i1, i2;

 for(i=0; i<*dimAlpha; i++)
    { ind = i + *dimBeta;
      alpha[i] = theta[ind];    }
   getRt(alpha, *dimAlpha, Rt);

  for(k=0; k<*dimQ; k++)
        v[k]= qnorm((nod[k]+1)/2,0,1,1,0);
   
  *l=0.0;
  for(i=0; i<*dimData; i++){
    etab[i]=0.0; 
    for(j=0; j< *dimBeta; j++)
      etab[i]+=MAT2(X,i,j,*dimData) * theta[j];}
  ind= 0;

  for(i=0; i<*dimG; i++)
    { n = nG[i];
      r = 0.0;
      for(i1=0; i1<*dimQ; i1++)
       for(i2=0; i2<*dimQ; i2++)
	{ A=0.0;
	  uv[0]=v[i1];
	  uv[1]=v[i2];  
	  matprod(Rt,uv,2,2,2,1,uvR);
	 for(j=0; j<n; j++)
	  { k = j + ind;
	    yM =  y[k];      
	    denM = den[k];
	    u = 0.0; 
	    for(z=0; z<2; z++)
	      u += MAT2(Z,k,z,*dimData) * uvR[z];
 	    eta = etab[k] + u;
	    prob = exp(eta) / (1.0 + exp(eta));
	    A += yM * log(prob)+ (denM - yM) * log(1.0-prob);}
	    r +=  w[i1] * w[i2] * 0.25 * exp(A);}
      ind += nG[i];
      (*l) += log(r);    }
};


void grad_GH2(double *theta,
       	     int *dimBeta,
	     int *dimAlpha,
	     double *X,
	     double *Z, 
	     int *y,
	     int *den,
	     int *dimData,
             int *nG,
	     int *dimG,
	     int *maxG,
	     double *nod,
	     double *w,
	     int *dimQ,
             double *grad)
{ 
  double *etab=(double *) R_alloc(*dimData, sizeof(double));
  double *alpha=(double *) R_alloc(*dimAlpha, sizeof(double));
  double *v=(double *) R_alloc(*dimQ, sizeof(double));
  double *Az=(double *) R_alloc(*dimBeta, sizeof(double));
  double *rp=(double *) R_alloc(*dimBeta, sizeof(double));
  double Rt[4] = {0.0, 0.0, 0.0, 0.0};
  double uv[2], uvR[2];
  double eta, prob, r, u, A;
  int i,j,ind,k,h,n, yM, denM, z, i1, i2;

 for(i=0; i<*dimAlpha; i++)
    { ind = i + *dimBeta;
      alpha[i] = theta[ind];    }
   getRt(alpha, *dimAlpha, Rt);

  for(k=0; k<*dimQ; k++)
        v[k]= qnorm((nod[k]+1)/2,0,1,1,0);
   
  for(i=0; i<*dimData; i++){
    etab[i]=0.0; 
    for(j=0; j< *dimBeta; j++)
      etab[i]+=MAT2(X,i,j,*dimData) * theta[j];}
  ind= 0;

  for(i=0; i<*dimG; i++)
    { n = nG[i];
      r = 0.0;
      iniz(rp, *dimBeta);
      for(i1=0; i1<*dimQ; i1++)
       for(i2=0; i2<*dimQ; i2++)
	{ A=0.0;
	  iniz(Az,*dimBeta);
	  uv[0]=v[i1];
	  uv[1]=v[i2];  
	  matprod(Rt,uv,2,2,2,1,uvR);
	 for(j=0; j<n; j++)
	  { k = j + ind;
	    yM =  y[k];      
	    denM = den[k];
	    u = 0.0; 
	    for(z=0; z<2; z++)
	      u += MAT2(Z,k,z,*dimData) * uvR[z];
 	    eta = etab[k] + u;
	    prob = exp(eta) / (1.0 + exp(eta));
	    A += yM * log(prob)+ (denM - yM) * log(1.0-prob);
            for(z=0; z<*dimBeta; z++)
	      (Az[z]) +=  (yM - prob *denM) * MAT2(X,k,z,*dimData);  }
	    r +=  w[i1] * w[i2] * 0.25 * exp(A);
            for(z=0; z<*dimBeta; z++)
              (rp[z]) +=   w[i1] * w[i2] * 0.25 * exp(A) * Az[z]; }
      ind += nG[i];
      MAT2(grad,i,*dimBeta, *dimG) = r;
      for( z=0; z<*dimBeta; z++)
	MAT2(grad,i,z,*dimG) = rp[z];
    }
};


/*  #####################################     */
/*         FUNCTIONS FOR TWO-PART CASE           */
/* #####################################      */ 

void lik_tp_GH1(double *theta,
       	     int *dimTheta,
	     int *dimAlpha,	
	     double *X,
	     int *dimBeta,	
	     double *Xs,
             int *dimGamma,	
	     int *y,
	     double *v,
	     int *dimData,
             int *nG,
	     int *dimG,
	     int *maxG, 
	     double *nod,
	     double *w,
	     int *dimQ,
             double *l)
{ double *etab=(double *) R_alloc(*dimData, sizeof(double));
  double *u=(double *) R_alloc(*dimQ, sizeof(double));
  double *csA=(double *) R_alloc(*dimQ, sizeof(double));
  double *alpha=(double *) R_alloc(*dimAlpha, sizeof(double));
  double *gamma=(double *) R_alloc(*dimGamma, sizeof(double));
  double Rt[4] = {0.0, 0.0, 0.0, 0.0};
  double Psi[4] = {0.0, 0.0, 0.0, 0.0};
  double R[4] = {0.0, 0.0, 0.0, 0.0};
  double eta, prob, r, s2, s2c, s2d, sdc, H,  a, mu, B, E, D, N, suma, suma2;
  int i,j,ind,k,h,n, sumy;
  
 for(i=0; i<*dimAlpha; i++)
    { ind = i + *dimBeta + *dimGamma;
      alpha[i] = theta[ind];    }
 for(i=0; i<*dimGamma; i++)
    { ind = i + *dimBeta;
      gamma[i] = theta[ind];    }
   getRt(alpha, *dimAlpha, Rt);
   trasp(Rt,2,2,R);
   matprod(Rt,R,2,2,2,2,Psi);
   s2c = MAT2(Psi,0,0,2);
   s2d = MAT2(Psi,1,1,2);
   sdc = MAT2(Psi,0,1,2);
   H = (s2d - sdc *sdc / s2c);
   ind = *dimTheta-1;
   s2 = theta[ind] * theta[ind];
   for(k=0; k<*dimQ; k++)
     u[k]= qnorm((nod[k]+1)/2,0,1,1,0) * sqrt(s2c);
   *l=0.0;
   for(i=0; i<*dimData; i++){
     etab[i]=0.0; 
     for(j=0; j< *dimBeta; j++)
       etab[i]+=MAT2(X,i,j,*dimData) * theta[j];}
   ind= 0;
   for(i=0; i<*dimG; i++)
     { n = nG[i];
       sumy = 0;
       suma = 0.0;
       suma2 =0.0;
       for(j=0; j<n; j++)
	 {k = j + ind;
          sumy += y[k];
	  mu = 0.0;
	  for(h=0; h< *dimGamma; h++)
              mu += MAT2(Xs,k,h,*dimData) * gamma[h];
	  a = (v[k] - mu) * y[k];
	  suma += a;
	  suma2 += a * a;
	 }
       B= sumy + s2 / H; 
       E= suma * (sdc/s2c) / (H * B);  
       D=  sumy * (sdc/s2c) * (sdc/s2c) / (H * B) ; 
       for(h=0; h<*dimQ; h++)
	 {csA[h]=0.0;
	   for(j=0; j<n; j++)
	   { k = j + ind;
	     eta = etab[k] + u[h];
	     prob = exp(eta) / (1.0 + exp(eta));
	     (csA[h]) += y[k] * log(prob)+ (1.0 - y[k]) * log(1.0-prob);}
	     } 
       r = 0.0;
       for(h=0; h<*dimQ; h++)
	 r += exp(csA[h]+  u[h] * E - 0.5 * u[h] *u[h] * D) * w[h] *0.5;
       ind += nG[i];
       N= -0.5 * log(H) -sumy * 0.5 * log(s2) + 0.5 * log(s2 /B) - 0.5*(suma2-suma*suma /B) /s2;
       (*l) += log(r)+N;
       }
 };


void grad_tp_GH1(double *theta,
       	     int *dimTheta,
	     int *dimAlpha,	
	     double *X,
	     int *dimBeta,	
	     double *Xs,
             int *dimGamma,	
	     int *y,
	     double *v,
	     int *dimData,
             int *nG,
	     int *dimG,
	     int *maxG, 
	     double *nod,
	     double *w,
	     int *dimQ,
             double *grad)
{ double *etab=(double *) R_alloc(*dimData, sizeof(double));
  double *a=(double *) R_alloc(*dimData, sizeof(double));
  double *u=(double *) R_alloc(*dimQ, sizeof(double));
  double *csA=(double *) R_alloc(*dimQ, sizeof(double));
  double *alpha=(double *) R_alloc(*dimAlpha, sizeof(double));
  double *gamma=(double *) R_alloc(*dimGamma, sizeof(double));
  double *csAz=(double *) R_alloc(*dimQ * *dimBeta, sizeof(double));
  double *rp=(double *) R_alloc(*dimBeta + *dimGamma, sizeof(double));
  double *sumxz=(double *) R_alloc(*dimGamma, sizeof(double));
  double *sumaxz=(double *) R_alloc(*dimGamma, sizeof(double));
  double Rt[4] = {0.0, 0.0, 0.0, 0.0};
  double Psi[4] = {0.0, 0.0, 0.0, 0.0};
  double R[4] = {0.0, 0.0, 0.0, 0.0};
  double eta, prob, r, s2, s2c, s2d, sdc, H,  mu, B, E, D, N, suma, suma2,Ed,np;
  int i,j,ind,k,h,n, sumy,p,z,z1;
  
 for(i=0; i<*dimAlpha; i++)
    { ind = i + *dimBeta + *dimGamma;
      alpha[i] = theta[ind];    }
 for(i=0; i<*dimGamma; i++)
    { ind = i + *dimBeta;
      gamma[i] = theta[ind];    }
   getRt(alpha, *dimAlpha, Rt);
   trasp(Rt,2,2,R);
   matprod(Rt,R,2,2,2,2,Psi);
   s2c = MAT2(Psi,0,0,2);
   s2d = MAT2(Psi,1,1,2);
   sdc = MAT2(Psi,0,1,2);
   H = (s2d - sdc *sdc / s2c);
   ind = *dimTheta-1;
   s2 = theta[ind] * theta[ind];
   for(k=0; k<*dimQ; k++)
     u[k]= qnorm((nod[k]+1)/2,0,1,1,0) * sqrt(s2c);
   for(i=0; i<*dimData; i++){
     etab[i]=0.0; 
     for(j=0; j< *dimBeta; j++)
       etab[i]+=MAT2(X,i,j,*dimData) * theta[j];
     mu=0.0; 
     for(j=0; j< *dimGamma; j++)
       mu += MAT2(Xs,i,j,*dimData) * gamma[j];
     a[i]= (v[i] - mu) * y[i];
   }
   ind= 0;
   for(i=0; i<*dimG; i++)
     { n = nG[i];
       sumy = 0;
       suma = 0.0;
       suma2 =0.0;
       iniz(sumxz, *dimGamma);
       iniz(sumaxz, *dimGamma);
       for(j=0; j<n; j++)
	 {k = j + ind;
          sumy += y[k];
	  for(h=0; h< *dimGamma; h++)
              { (sumxz[h]) +=  y[k] * MAT2(Xs,k,h,*dimData);
		(sumaxz[h]) += a[k] * MAT2(Xs,k,h,*dimData); }
	  suma += a[k];
	  suma2 += a[k] * a[k];
	 }
       B= sumy + s2 / H; 
       E= suma * (sdc/s2c) / (H * B);  
       D=  sumy * (sdc/s2c) * (sdc/s2c) / (H * B) ; 
       for(h=0; h<*dimQ; h++)
	 {csA[h]=0.0;
	  for( z=0; z<*dimBeta; z++)
	      MAT2(csAz, z, h,*dimBeta)=0.0;
	   for(j=0; j<n; j++)
	   { k = j + ind;
	     eta = etab[k] + u[h];
	     prob = exp(eta) / (1.0 + exp(eta));
	     (csA[h]) += y[k] * log(prob)+ (1.0 - y[k]) * log(1.0-prob);
             for( z=0; z<*dimBeta; z++)
                 MAT2(csAz,z,h,*dimBeta) += (y[k]-prob) * MAT2(X,k,z,*dimData) ;
	   }
	     } 
       r = 0.0;
      iniz(rp, *dimBeta + *dimGamma);
      for(h=0; h<*dimQ; h++)
	{ r += exp(csA[h] + u[h] * E - 0.5 * u[h] *u[h] * D) * w[h] *0.5;
          for( z=0; z<*dimBeta; z++)
            rp[z] +=  exp(csA[h] + u[h] * E - 0.5 * u[h] *u[h] * D) * MAT2(csAz,z,h,*dimBeta) * w[h] * 0.5;
	  p = *dimBeta + *dimGamma;
          for( z=*dimBeta; z<p; z++)
	    { z1 = z - *dimBeta;
              Ed = (-1.0) * sumxz[z1] * (sdc/s2c) / (H * B);
	      rp[z] +=  exp(csA[h]+ u[h] * E - 0.5 * u[h] *u[h] * D) * u[h] *Ed * w[h] * 0.5; } 
	  }
      ind += nG[i];
      for( z=0; z<*dimBeta; z++)
	MAT2(grad,i,z,*dimG) = rp[z] / r;
      for( z=*dimBeta; z<p; z++)
	{ z1 = z - *dimBeta;
	  np =  sumaxz[z1] /s2 - suma * sumxz[z1] /(s2 * B);
  	  MAT2(grad,i,z,*dimG) = rp[z] / r + np;}
     }
 };




